/*
 * Author: Evan Peterson
 * Cal Poly, CPE-471, Winter 2014
 */

#include <math.h>

#include "geom.h"

#include "board.h"
#include "tree.h"
#include "ground.h"
#include "gate.h"

#include "third_party/GeometryCreator.h"

RenderingHelper g_mStack;

// Shader attributes.
GLint h_aPosition;
GLint h_aNormal;
GLint h_aTexCoord;

// Shader uniforms.
GLint h_uModelMx;
GLint h_uMatAmb, h_uMatDif, h_uMatSpec, h_uMatShine;

// Simple generated 3D meshes.
static Mesh *g_meshes[MAX_MESH_INDEX + 1];

static Board *g_board;
static Ground *g_ground;
static std::vector<Tree *> g_trees;
static std::vector<Gate *> g_gates;

static int g_numTrees = 100;
static int g_numGates = 3;

int g_boostPoints = 0;
const int g_boostReward = 100;
const int g_boostPointsMax = 200;

float g_boardAccel = 0;
const float g_boardAccelMax = .02;
const float g_boardAccelMin = -g_boardAccelMax / 2;
const float g_boardAccelRate = .001;

const float g_boardVelMax = 2.0;
const float g_boardVelMin = .5;
float g_boardVel = g_boardVelMin;

Model::Model(const int meshIndex)
{
   // Load the model from a generated primitive model.
   GeometryCreator::loadIndexBufferObjectFromMesh(g_meshes[meshIndex],
    &this->polyCount, &this->vbo, &this->nbo, &this->tbo, &this->ibo);
}

void Model::draw(void)
{
   // Bind model matrix uniform.
   safe_glUniformMatrix4fv(h_uModelMx,
    glm::value_ptr(g_mStack.modelViewMatrix));

   // Bind vertex attributes.
   glBindBuffer(GL_ARRAY_BUFFER, this->vbo);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ARRAY_BUFFER, this->nbo);
   safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);
   glBindBuffer(GL_ARRAY_BUFFER, this->tbo);
   safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0);

   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, this->ibo);

   // Draw geometry
   glDrawElements(GL_TRIANGLES, this->polyCount * 3, GL_UNSIGNED_INT, 0);
}

void drawGeom()
{
   bool visible = false;

   // Draw models.
   g_board->draw();
   g_ground->draw();

   // Iterate over and draw trees.
   std::vector<Tree *>::iterator tree = g_trees.begin();
   while (tree != g_trees.end()) {
      visible = (*tree)->draw();
      if (!visible) {
         (*tree)->init();
      }
      tree++;
   }

   // Iterate over and draw gates.
   std::vector<Gate *>::iterator gate = g_gates.begin();
   while (gate != g_gates.end()) {
      visible = (*gate)->draw();
      (*gate)->checkHit(g_board->getPos());
      if (!visible) {
         (*gate)->init();
      }
      gate++;
   }
}

void initGeom()
{
   Gate *gate;

   // Generate some simple models.
   g_meshes[CYLINDER_MESH] = GeometryCreator::CreateCylinder(
    1.f, 1.f, 1.5f, 3, 3);
   g_meshes[CONE_MESH] = GeometryCreator::CreateCylinder(
    .8f, 0.f, 1.6f, 8, 16);
   g_meshes[GROUND_MESH] = GeometryCreator::CreateCylinder(
    1.f, 1.f, 1.5f, 8, 8);
   g_meshes[CUBE_MESH] = GeometryCreator::CreateCube(glm::vec3(1.f, 1.f, 1.f));
   g_meshes[PLANE_MESH] = GeometryCreator::CreatePlane(
    glm::vec3(1.f, 1.f, 1.f));
   g_meshes[SPHERE_MESH] = GeometryCreator::CreateSphere(glm::vec3(1.5f));
   g_meshes[SMOOTH_CYL_MESH] = GeometryCreator::CreateCylinder(
    1.f, 1.f, 1.5f, 4, 8);


   // Set up a bunch of trees!
   while (g_trees.size() < g_numTrees) {
      g_trees.push_back(new Tree());
   }

   // Set up a few gates.
   while (g_gates.size() < g_numGates) {
      gate = new Gate();
      g_gates.push_back(gate);
      gate->setZ(g_gates.size());
   }

   g_board = new Board();
   g_ground = new Ground();
}

void setMaterial(int matId)
{
   float charge = g_boostPoints / (float) g_boostPointsMax;

   safe_glUniform3f(h_uMatAmb, 0.1, 0.1, 0.1);

   switch(matId) {
   case BOARD_MAT:
      safe_glUniform3f(h_uMatDif, 0.2, 1.0 * charge, 0.2);
      safe_glUniform3f(h_uMatSpec, 0.8, 0.8, 0.8);
      safe_glUniform1f(h_uMatShine, 10.0);
      break;
   case PINE_MAT:
      safe_glUniform3f(h_uMatDif, 0.2, 0.2, 0.4);
      safe_glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1);
      safe_glUniform1f(h_uMatShine, 1.0);
      break;
   case BARK_MAT:
      safe_glUniform3f(h_uMatDif, 0.6, 0.2, 0.2);
      safe_glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1);
      safe_glUniform1f(h_uMatShine, 1.0);
      break;
   case SNOW_MAT:
      safe_glUniform3f(h_uMatDif, 0.1, 0.1, 0.3);
      safe_glUniform3f(h_uMatSpec, 0.1, 0.1, 0.4);
      safe_glUniform1f(h_uMatShine, 5.0);
      break;
   case TUX_MAT:
      safe_glUniform3f(h_uMatDif, 0.1, 0.1, 0.1);
      safe_glUniform3f(h_uMatSpec, 0.5, 0.5, 0.5);
      safe_glUniform1f(h_uMatShine, 100.0);
      break;
   case ORANGE_MAT:
      safe_glUniform3f(h_uMatDif, 0.95, 0.15, 0.4);
      safe_glUniform3f(h_uMatSpec, 0.5, 0.15, 0.35);
      safe_glUniform1f(h_uMatShine, 50.0);
      safe_glUniform3f(h_uMatAmb, 0.4, 0.1, 0.2);
      break;
   }
}
