/*
 * Author: Evan Peterson
 * Cal Poly, CPE-471, Winter 2014
 */

#include "geom.h"
#include "main.h"
#include "texture.h"

#include <vector>
#include <unistd.h> // usleep

// Prototypes -- so I can hide these functions at the bottom of the file.
int installShader(const GLchar *vShaderName, const GLchar *fShaderName);
float p2wx(int in_x);
float p2wy(int in_y);

// Shader uniforms.
GLint h_uViewMx, h_uProjMx;
GLint h_uTexUnit;
GLint h_uCamPos;
GLint h_uLightColor, h_uLightDir;

static int g_shadeProg;
static float g_height = 600;
static float g_width = 800;
float g_viewBound = 5;

// Camera information.
static glm::vec3 g_lookAt(0, .667, -1);
static glm::vec3 g_camTrans(0, -4, -5);

// Mouse information.
static glm::vec2 g_mouseDown(0);
static bool g_mouseDrag = false;

// Joystick information.
bool g_usingJoystick = false;
struct joystick g_js;

// Controls
int g_doBoost = 0;
float g_axisValue = 0;

// Set the projection matrix in the shaders.
void setProjection(void)
{
   glm::mat4 proj = glm::perspective(80.0f, g_width / g_height,
    0.1f, 300.f);
   safe_glUniformMatrix4fv(h_uProjMx, glm::value_ptr(proj));
}

// Set the view matrix in the shaders.
void setView(void)
{
   glm::mat4 eye = glm::lookAt(glm::vec3(0), g_lookAt, glm::vec3(0, 1, 0));
   glm::mat4 trans = glm::translate(glm::mat4(1), g_camTrans);
   glm::mat4 view = trans * eye;

   safe_glUniformMatrix4fv(h_uViewMx, glm::value_ptr(view));
   safe_glUniform3f(h_uCamPos, g_camTrans.x, g_camTrans.y, g_camTrans.z);
}

void setLight(void)
{
   safe_glUniform3f(h_uLightDir, -1.f, 1.f, 1.f); // Light is up-back-and-left.
   safe_glUniform3f(h_uLightColor, 1.f, 1.f, 1.f); // Simple white light.
}

// GLFW callback for when the window is resized.
void glfwWindowResize(GLFWwindow *window, int width, int height)
{
   g_width = (float)width;
   g_height = (float)height;
   glViewport(0, 0, (GLsizei)(width), (GLsizei)(height));
}

// GLFW main drawing routine.
void glfwDraw(GLFWwindow *window)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   // Enable the shader.
   glUseProgram(g_shadeProg);

   // Enable the texture unit.
   glEnable(GL_TEXTURE_2D);
   glActiveTexture(GL_TEXTURE0);
   safe_glUniform1i(h_uTexUnit, 0);


   // Enable attributes used by the shader.
   safe_glEnableVertexAttribArray(h_aPosition);
   safe_glEnableVertexAttribArray(h_aNormal);
   safe_glEnableVertexAttribArray(h_aTexCoord);

   // Send View and Projection matrix to the shader.
   setProjection();
   setView();
   setLight();

   // Draw the geometry.
   drawGeom();

   // Disable attributes used by the shader.
   safe_glDisableVertexAttribArray(h_aPosition);
   safe_glDisableVertexAttribArray(h_aNormal);
   safe_glDisableVertexAttribArray(h_aTexCoord);

   // Disable the shader.
   glUseProgram(0);
   glfwSwapBuffers(window);

   // 1 second / 70 microseconds
   usleep(1000000.f / 70.f);
}

// GLFW callback for whenever a mouse button is pressed/released.
void glfwClick(GLFWwindow *window, int button, int action, int mods)
{
   double screenX, screenY;
   if (action == GLFW_PRESS) {
      g_mouseDrag = true;
      glfwGetCursorPos(window, &screenX, &screenY);
      g_mouseDown.x = p2wx(screenX);
      g_mouseDown.y = p2wy(screenY);
   }
   else if (action == GLFW_RELEASE) {
      g_mouseDrag = false;
   }
}

// GLFW callback for whenever the mouse cursor is moved.
void glfwDrag(GLFWwindow *window, double x, double y)
{
   static glm::vec2 mousePos(0);

   if (g_mouseDrag) {
      mousePos.x = p2wx(x);
      mousePos.y = p2wy(y);

      // Update the last mouse position to be the current position.
      g_mouseDown.x = mousePos.x;
      g_mouseDown.y = mousePos.y;
   }
}

// GLFW callback for whenever a key is pressed.
void glfwKeyPress(GLFWwindow *window, int key, int scan, int action, int mods)
{
   if (key == GLFW_KEY_W && action == GLFW_RELEASE) {
      g_doBoost = false;
   }

   if (action == GLFW_PRESS || action == GLFW_REPEAT) {
      switch (key) {
      // Board controls (lame, use a controller, please.)
      case GLFW_KEY_W:
         g_doBoost = true;
         break;
      case GLFW_KEY_A:
         g_axisValue = fmax(-1, g_axisValue - .2);
         break;
      case GLFW_KEY_D:
         g_axisValue = fmin(1, g_axisValue + .2);
         break;
      // Camera controls.
      case GLFW_KEY_I:
         g_camTrans.z += .3;
         break;
      case GLFW_KEY_J:
         g_camTrans.x += .3;
         break;
      case GLFW_KEY_K:
         g_camTrans.z -= .3;
         break;
      case GLFW_KEY_L:
         g_camTrans.x -= .3;
         break;
      // Get out of here, man!
      case GLFW_KEY_ESCAPE:
         glfwSetWindowShouldClose(window, GL_TRUE);
         break;
      }
   }
}

// GLFW callback for any detected errors.
void glfwError(int error, const char *description)
{
   std::cerr << description << std::endl;
}

// Perform some initial GL setup.
void glInitialize(void)
{
   // TODO GL_VERSION not being set?
   //getGLversion();
   srand(time(0));

   glClearColor(0.7f, 0.8f, 0.9f, 1); // Set the default pixel color.
   glClearDepth(1); // Depth buffer setup.
   glDepthFunc(GL_LEQUAL); // Type of depth testing.
   glEnable(GL_DEPTH_TEST); // Enable depth testing.
   glEnable(GL_CULL_FACE); // Turn on backface culling.

   g_mStack.useModelViewMatrix();
   g_mStack.loadIdentity();

   // Install the shader.
   if (!installShader(textFileRead((char *)"vert.glsl"),
                      textFileRead((char *)"frag.glsl"))) {
      std::cout << "Error installing shader!" << std::endl;
      exit(EXIT_FAILURE);
   }

   initGeom();
   initTextures();
}

bool initJoystick(GLFWwindow *window)
{
   int jsId;

   // Look for a connected joystick.
   g_js.id = -1;
   for (jsId = GLFW_JOYSTICK_1; jsId < GLFW_JOYSTICK_LAST + 1; jsId++) {
      if (glfwJoystickPresent(jsId)) {
         g_js.id = jsId;
         g_usingJoystick = true;
         break;
      }
   }

   if (g_js.id < 0) {
      std::cerr << "No joystick was found!" << std::endl;
      g_usingJoystick = false;
   }
}

void handleJoystick(GLFWwindow *window)
{
   if (!g_usingJoystick)
      return;

   if (!glfwJoystickPresent(g_js.id)) {
      std::cerr << "Lost the joystick!" << std::endl;
      g_usingJoystick = false;
   }
   else {
      g_js.axes = glfwGetJoystickAxes(g_js.id, &g_js.axisCount);
      g_js.buttons = glfwGetJoystickButtons(g_js.id, &g_js.buttonCount);
      g_doBoost = g_js.buttons[A_BTN];
      g_axisValue = g_js.axes[LEFT_X];
   }
}

int main(int argc, char *argv[])
{
   /* GLFW Setup */
   GLFWwindow* window;

   glfwSetErrorCallback(glfwError);
   if (!glfwInit()) {
      exit(EXIT_FAILURE);
   }

   // Set up the main window.
   window = glfwCreateWindow(g_width, g_height, "Carving", NULL, NULL);
   if (!window) {
      glfwTerminate();
      exit(EXIT_FAILURE);
   }
   glfwMakeContextCurrent(window);
   glfwSetWindowPos(window, 80, 80);

   // Set up event listeners.
   glfwSetWindowSizeCallback(window, glfwWindowResize);
   glfwSetKeyCallback(window, glfwKeyPress);
   //glfwSetMouseButtonCallback(window, glfwClick);
   //glfwSetCursorPosCallback(window, glfwDrag);

   // Other initializations.
   glInitialize();
   initJoystick(window);

   // Start the main execution loop.
   while (!glfwWindowShouldClose(window)) {
      handleJoystick(window);
      glfwPollEvents();
      glfwDraw(window);
   }

   // Clean up after GLFW.
   glfwDestroyWindow(window);
   glfwTerminate();
   exit(EXIT_SUCCESS);
}

// Convert pixel x-coordinate to world x-coordinate.
float p2wx(int in_x) {
   double left, right, scale, translate;

   right = g_width > g_height ? g_width / g_height : 1;
   left = -right;

   scale = (-g_width + 1) / (left - right);
   translate = scale * -left;

   return (in_x - translate) / scale;
}

// Convert pixel y-coordinate to world y-coordinate.
float p2wy(int in_y) {
   double bottom, top, scale, translate;

   // NOTE: No need to flip glfw y-coordinate.
   top = g_height >= g_width ? g_height / g_width : 1;
   bottom = -top;

   scale = (-g_height + 1) / (bottom - top);
   translate = scale * -bottom;

   return (in_y - translate) / scale;
}

// Function to help load vertex and fragment shaders.
int installShader(const GLchar *vShaderName, const GLchar *fShaderName) {
   GLuint VS; //handles to shader object
   GLuint FS; //handles to frag shader object
   GLint vCompiled, fCompiled, linked; //status of shader

   VS = glCreateShader(GL_VERTEX_SHADER);
   FS = glCreateShader(GL_FRAGMENT_SHADER);

   //load the source
   glShaderSource(VS, 1, &vShaderName, NULL);
   glShaderSource(FS, 1, &fShaderName, NULL);

   //compile shader and print log
   glCompileShader(VS);
   /* check shader status requires helper functions */
   printOpenGLError();
   glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
   printShaderInfoLog(VS);

   //compile shader and print log
   glCompileShader(FS);
   /* check shader status requires helper functions */
   printOpenGLError();
   glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
   printShaderInfoLog(FS);

   if (!vCompiled || !fCompiled) {
      printf("Error compiling either shader %s or %s", vShaderName, fShaderName);
      return 0;
   }

   //create a program object and attach the compiled shader
   g_shadeProg = glCreateProgram();
   glAttachShader(g_shadeProg, VS);
   glAttachShader(g_shadeProg, FS);

   glLinkProgram(g_shadeProg);
   /* check shader status requires helper functions */
   printOpenGLError();
   glGetProgramiv(g_shadeProg, GL_LINK_STATUS, &linked);
   printProgramInfoLog(g_shadeProg);

   glUseProgram(g_shadeProg);

   /* get handles to attribute and uniform data in shader */
   h_aPosition = safe_glGetAttribLocation(g_shadeProg, "aPosition");
   h_aNormal = safe_glGetAttribLocation(g_shadeProg, "aNormal");
   h_aTexCoord = safe_glGetAttribLocation(g_shadeProg, "aTexCoord");
   h_uProjMx = safe_glGetUniformLocation(g_shadeProg, "uProjMatrix");
   h_uViewMx = safe_glGetUniformLocation(g_shadeProg, "uViewMatrix");
   h_uModelMx = safe_glGetUniformLocation(g_shadeProg, "uModelMatrix");
   h_uTexUnit = safe_glGetUniformLocation(g_shadeProg, "uTexUnit");
   h_uLightDir = safe_glGetUniformLocation(g_shadeProg, "uLightDir");
   h_uLightColor = safe_glGetUniformLocation(g_shadeProg, "uLightColor");
   /* material data */
   h_uMatAmb = safe_glGetUniformLocation(g_shadeProg, "uMat.aColor");
   h_uMatDif = safe_glGetUniformLocation(g_shadeProg, "uMat.dColor");
   h_uMatSpec = safe_glGetUniformLocation(g_shadeProg, "uMat.sColor");
   h_uMatShine = safe_glGetUniformLocation(g_shadeProg, "uMat.shine");

   printf("Successfully installed shader %d.\n", g_shadeProg);
   return 1;
}
